"""
Python library calling enry Go implementation trough cFFI (API, out-of-line) and Cgo.
"""
from typing import List

from _c_enry import lib

from enry.types import Guess
from enry.utils import go_str_slice_to_py, init_go_slice, py_bytes_to_go, py_str_to_go, transform_types, \
    transform_types_ret_str_slice

GetLanguage = transform_types([str, bytes], str)(lib.GetLanguage)
GetLanguageByContent = transform_types([str, bytes], Guess)(lib.GetLanguageByContent)
GetLanguageByExtension = transform_types([str], Guess)(lib.GetLanguageByExtension)
GetLanguageByFilename = transform_types([str], Guess)(lib.GetLanguageByFilename)
GetLanguageByModeline = transform_types([bytes], Guess)(lib.GetLanguageByModeline)
GetLanguageByShebang = transform_types([bytes], Guess)(lib.GetLanguageByShebang)
GetLanguageByEmacsModeline = transform_types([bytes], Guess)(lib.GetLanguageByEmacsModeline)
GetLanguageByVimModeline = transform_types([bytes], Guess)(lib.GetLanguageByVimModeline)

GetLanguages = transform_types_ret_str_slice([str, bytes])(lib.GetLanguages)

GetMimeType = transform_types([str, str], str)(lib.GetMimeType)
GetColor = transform_types([str], str)(lib.GetColor)

# TODO: GetLanguages
# TODO: GetLanguageExtensions
IsVendor = transform_types([str], bool)(lib.IsVendor)
IsGenerated = transform_types([str, bytes], bool)(lib.IsGenerated)
IsBinary = transform_types([bytes], bool)(lib.IsBinary)
IsConfiguration = transform_types([str], bool)(lib.IsConfiguration)
IsDocumentation = transform_types([str], bool)(lib.IsDocumentation)
IsDotFile = transform_types([str], bool)(lib.IsDotFile)
IsImage = transform_types([str], bool)(lib.IsImage)


def get_language(filename: str, content: bytes) -> str:
    return GetLanguage(filename, content)


def get_language_by_content(filename: str, content: bytes) -> Guess:
    return GetLanguageByContent(filename, content)


def get_language_by_extension(filename: str) -> Guess:
    return GetLanguageByExtension(filename)


def get_language_by_filename(filename: str) -> Guess:
    return GetLanguageByFilename(filename)


def get_language_by_modeline(content: bytes) -> Guess:
    return GetLanguageByModeline(content)


def get_language_by_vim_modeline(content: bytes) -> Guess:
    return GetLanguageByVimModeline(content)


def get_language_by_emacs_modeline(content: bytes) -> Guess:
    return GetLanguageByEmacsModeline(content)


def get_language_by_shebang(content: bytes) -> Guess:
    return GetLanguageByShebang(content)


def get_languages(filename: str, content: bytes) -> List[str]:
    return GetLanguages(filename, content)


def get_mime_type(path: str, language: str) -> str:
    return GetMimeType(path, language)


def get_color(language: str) -> str:
    return GetColor(language)


def is_vendor(filename: str) -> bool:
    return IsVendor(filename)


def is_generated(filename: str, content: bytes) -> bool:
    return IsGenerated(filename, content)


def is_binary(content: bytes) -> bool:
    return IsBinary(content)


def is_configuration(path: str) -> bool:
    return IsConfiguration(path)


def is_documentation(path: str) -> bool:
    return IsDocumentation(path)


def is_dot_file(path: str) -> bool:
    return IsDotFile(path)


def is_image(path: str) -> bool:
    return IsImage(path)



from enry.definitions import get_color, get_language_by_shebang, get_language, get_language_by_content, \
    get_language_by_emacs_modeline, get_language_by_extension, get_language_by_filename, get_language_by_modeline, \
    get_language_by_vim_modeline, get_languages, get_mime_type, is_binary, is_configuration, is_documentation, \
    is_dot_file, \
    is_generated, is_image, is_vendor

__all__ = [
    "get_color",
    "get_language",
    "get_languages",
    "get_mime_type",
    "get_language_by_vim_modeline",
    "get_language_by_extension",
    "get_language_by_content",
    "get_language_by_emacs_modeline",
    "get_language_by_modeline",
    "get_language_by_filename",
    "get_language_by_shebang",
    "is_vendor",
    "is_binary",
    "is_image",
    "is_generated",
    "is_documentation",
    "is_dot_file",
    "is_configuration",
]



from enry import *

import pytest


@pytest.mark.parametrize("filename,content,language", [
    ("test.py", "import os", "Python"),
    ("", "#!/usr/bin/bash", "Shell"),
    ("test.hs", "", "Haskell"),
])
def test_get_language(filename: str, content: str, language: str):
    assert get_language(filename, content.encode()) == language


def test_get_language_by_filename():
    assert get_language_by_filename("pom.xml").language == "Maven POM"


def test_get_language_by_content():
    assert get_language_by_content("test.php", "<?php $foo = bar();".encode()).language == "PHP"


def test_get_language_by_emacs_modeline():
    modeline = "// -*- font:bar;mode:c++ -*-\ntemplate <typename X> class { X i; };"
    assert get_language_by_emacs_modeline(modeline.encode()).language == "C++"


def test_get_language_by_vim_modeline():
    modeline = "# vim: noexpandtab: ft=javascript"
    assert get_language_by_vim_modeline(modeline.encode()).language == "JavaScript"


@pytest.mark.parametrize("modeline,language", [
    ("// -*- font:bar;mode:c++ -*-\ntemplate <typename X> class { X i; };", "C++"),
    ("# vim: noexpandtab: ft=javascript", "JavaScript")
])
def test_get_language_by_modeline(modeline: str, language: str):
    assert get_language_by_modeline(modeline.encode()).language == language


def test_get_language_by_extension():
    assert get_language_by_extension("test.lisp").language == "Common Lisp"


def test_get_language_by_shebang():
    assert get_language_by_shebang("#!/usr/bin/python3".encode()).language == "Python"


def test_get_mime_type():
    assert get_mime_type("test.rb", "Ruby") == "text/x-ruby"


def test_is_binary():
    assert is_binary("println!('Hello world!\n');".encode()) == False


@pytest.mark.parametrize("path,is_documentation_actual", [
    ("sss/documentation/", True),
    ("docs/", True),
    ("test/", False),
])
def test_is_documentation(path: str, is_documentation_actual: bool):
    assert is_documentation(path) == is_documentation_actual


@pytest.mark.parametrize("path,is_dot_actual", [
    (".env", True),
    ("something.py", False),
])
def test_is_dot(path: str, is_dot_actual: bool):
    assert is_dot_file(path) == is_dot_actual


@pytest.mark.parametrize("path,is_config_actual", [
    ("configuration.yml", True),
    ("some_code.py", False),
])
def test_is_configuration(path: str, is_config_actual: bool):
    assert is_configuration(path) == is_config_actual


@pytest.mark.parametrize("path,is_image_actual", [
    ("nsfw.jpg", True),
    ("shrek-picture.png", True),
    ("openjdk-1000.parquet", False),
])
def test_is_image(path: str, is_image_actual: bool):
    assert is_image(path) == is_image_actual


def test_get_color():
    assert get_color("Go") == "#00ADD8"


def test_get_languages():
    assert get_languages("test.py", "import os".encode())


from functools import wraps
from typing import Hashable, List, Sequence

from _c_enry import ffi

from enry.types import Guess


def py_bytes_to_go(py_bytes: bytes):
    c_bytes = ffi.new("char[]", py_bytes)
    go_slice = ffi.new("GoSlice *", [c_bytes, len(py_bytes), len(py_bytes)])
    return (go_slice[0], c_bytes)


def py_str_to_go(py_str: str):
    str_bytes = py_str.encode()
    c_str = ffi.new("char[]", str_bytes)
    go_str = ffi.new("_GoString_ *", [c_str, len(str_bytes)])
    return (go_str[0], c_str)


def go_str_to_py(go_str: str):
    str_len = go_str.n
    if str_len > 0:
        return ffi.unpack(go_str.p, go_str.n).decode()
    return ""


def init_go_slice():
    return ffi.new("GoSlice *")


def go_str_slice_to_py(str_slice) -> List[str]:
    slice_len = str_slice.len
    char_arr = ffi.cast("char **", str_slice.data)
    return [ffi.string(char_arr[i]).decode() for i in range(slice_len)]


def go_bool_to_py(go_bool: bool):
    return go_bool == 1


def go_guess_to_py(guess) -> Guess:
    return Guess(go_str_to_py(guess.r0), go_bool_to_py(guess.r1))


py_to_go = {
    str: py_str_to_go,
    bytes: py_bytes_to_go,
}


go_to_py = {
    str: go_str_to_py,
    bool: go_bool_to_py,
    Guess: go_guess_to_py,
}


def transform_types(in_types: Sequence[Hashable], out_type: Hashable):
    def decorator(fn):
        @wraps(fn)
        def inner(*args):
            args_transformed = [py_to_go[type_](arg) for type_, arg in zip(in_types, args)]
            return go_to_py[out_type](fn(*(arg[0] for arg in args_transformed)))
        return inner
    return decorator


def transform_types_ret_str_slice(in_types: Sequence[Hashable]):
    def decorator(fn):
        @wraps(fn)
        def inner(*args):
            ret_slice = init_go_slice()
            args_transformed = [py_to_go[type_](arg) for type_, arg in zip(in_types, args)]
            fn(*(arg[0] for arg in args_transformed), ret_slice)
            return go_str_slice_to_py(ret_slice)
        return inner
    return decorator
